"""Components of the Tag model.  Stem + Optional Suffix = Tag."""
from enum import Enum

# This is here to avoid a circular import in parsers.time
EXP_TIME_ROUND_DIGITS: int = 6


class StemName(str, Enum):
    """Controlled list of Tag Stems."""

    output = "OUTPUT"
    input = "INPUT"
    intermediate = "INTERMEDIATE"
    input_dataset = "INPUT_DATASET"
    frame = "FRAME"
    movie = "MOVIE"
    stokes = "STOKES"
    movie_frame = "MOVIE_FRAME"
    task = "TASK"
    cs_step = "CS_STEP"
    modstate = "MODSTATE"
    dsps_repeat = "DSPS_REPEAT"
    calibrated = "CALIBRATED"  # A flag to indicate the data has been calibrated but not yet output
    quality = "QUALITY"
    exposure_time = "EXP_TIME"
    quality_task = "QUALITY_TASK"


class Tag:
    """Controlled methods for creating tags from stems + optional suffixes."""

    @staticmethod
    def format_tag(stem: StemName | str, *parts):
        """
        Create a formatted tag sting given the input parts.

        Parameters
        ----------
        stem
            The name of the stem
        parts
            The remaining tag parts
        Returns
        -------
        The concatenated tag name
        """
        if isinstance(stem, Enum):
            stem = stem.value
        parts = [stem, *parts]
        return "_".join([str(part).upper() for part in parts])

    # Static Tags
    @classmethod
    def movie_frame(cls):
        """
        Return a movie frame tag.

        Returns
        -------
        A movie frame tag
        """
        return cls.format_tag(StemName.movie_frame)

    @classmethod
    def input(cls):
        """
        Return an input tag.

        Returns
        -------
        An input tag
        """
        return cls.format_tag(StemName.input)

    @classmethod
    def calibrated(cls) -> str:
        """
        Return a calibrated tag.

        Returns
        -------
        A calibrated tag
        """
        return cls.format_tag(StemName.calibrated)

    @classmethod
    def output(cls):
        """
        Return an output tag.

        Returns
        -------
        An output tag
        """
        return cls.format_tag(StemName.output)

    @classmethod
    def frame(cls):
        """
        Return a frame tag.

        Returns
        -------
        A frame tag
        """
        return cls.format_tag(StemName.frame)

    @classmethod
    def intermediate(cls):
        """
        Return an intermediate tag.

        Returns
        -------
        An intermediate tag
        """
        return cls.format_tag(StemName.intermediate)

    @classmethod
    def input_dataset_observe_frames(cls):
        """
        Return an input dataset observe frames tag.

        Returns
        -------
        An input dataset observe frames tag
        """
        return cls.format_tag(StemName.input_dataset, "observe_frames")

    @classmethod
    def input_dataset_calibration_frames(cls):
        """
        Return an input dataset calibration frames tag.

        Returns
        -------
        An input dataset calibration frames tag
        """
        return cls.format_tag(StemName.input_dataset, "calibration_frames")

    @classmethod
    def input_dataset_parameters(cls):
        """
        Return an input dataset parameters tag.

        Returns
        -------
        An input dataset parameters tag
        """
        return cls.format_tag(StemName.input_dataset, "parameters")

    @classmethod
    def movie(cls):
        """
        Return a movie tag.

        Returns
        -------
        A movie tag
        """
        return cls.format_tag(StemName.movie)

    # Dynamic Tags
    @classmethod
    def task(cls, ip_task_type: str):
        """
        Return a task tag for the given task type.

        Parameters
        ----------
        ip_task_type
            The task type
        Returns
        -------
        A task tag for the given type
        """
        return cls.format_tag(StemName.task, ip_task_type)

    @classmethod
    def cs_step(cls, n: int):
        """
        Return a cs step tag for the given cs_step number.

        Parameters
        ----------
        n
            The cs Step number

        Returns
        -------
        A cs Step tag for the given CS number
        """
        return cls.format_tag(StemName.cs_step, n)

    @classmethod
    def modstate(cls, n: int):
        """
        Return a modstate tag for the given modstate number.

        Parameters
        ----------
        n
            The modstate number

        Returns
        -------
        A modstate tag for the given modstate number
        """
        return cls.format_tag(StemName.modstate, n)

    @classmethod
    def stokes(cls, stokes_state: str) -> str:
        """
        Return a stokes tag for the given stokes value (I, Q, U, V).

        Parameters
        ----------
        stokes_state
            The input stokes state

        Returns
        -------
        A stokes tag for the given stokes state
        """
        return cls.format_tag(StemName.stokes, stokes_state)

    @classmethod
    def dsps_repeat(cls, dsps_repeat_number: int):
        """
        Return a dsps repeat tag for the given dsps_repeat number.

        Parameters
        ----------
        dsps_repeat_number
            The dsps repeat number

        Returns
        -------
        A dsps Repeat tag for the given dsps repeat number
        """
        return cls.format_tag(StemName.dsps_repeat, dsps_repeat_number)

    @classmethod
    def quality(cls, quality_metric: str) -> str:
        """
        Return a quality tag for the given quality metric.

        Parameters
        ----------
        quality_metric
            The input quality metric

        Returns
        -------
        A quality tag for the given quality metric
        """
        return cls.format_tag(StemName.quality, quality_metric)

    @classmethod
    def exposure_time(cls, exposure_time_s: float) -> str:
        """
        Return an exposure time tag for the given exposure time.

        Parameters
        ----------
        exposure_time_s
            The exposure time in seconds
        Returns
        -------
        An exposure time tag for the given exposure time
        """
        return cls.format_tag(
            StemName.exposure_time, round(float(exposure_time_s), EXP_TIME_ROUND_DIGITS)
        )

    @classmethod
    def quality_task(cls, quality_task_type: str) -> str:
        """
        Return a quality task tag for the given quality task type.

        Parameters
        ----------
        quality_task_type

        Returns
        -------
        A quality task tag for the given quality task type
        """
        return cls.format_tag(StemName.quality_task, quality_task_type)



from dataclasses import asdict
from dataclasses import dataclass
from pathlib import Path

import pytest

from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks import TransferL1Data
from dkist_processing_common.tests.conftest import FakeGQLClient


@dataclass
class FakeConstantsDb:
    PROPOSAL_ID: str = "PROPID"


def fake_list_objects(self, bucket, prefix=None):
    return list()


@pytest.fixture
def transfer_l1_data_task(recipe_run_id, tmp_path):
    with TransferL1Data(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        task.scratch = WorkflowFileSystem(
            recipe_run_id=recipe_run_id,
            scratch_base_path=tmp_path,
        )
        task.constants._update(asdict(FakeConstantsDb()))
        frame_path = task.scratch.workflow_base_path / Path("frame.fits")
        movie_path = task.scratch.workflow_base_path / Path("movie.mp4")
        with open(frame_path, "w") as f:
            f.write("Frame")
        task.tag(path=frame_path, tags=[Tag.frame(), Tag.output()])
        with open(movie_path, "w") as f:
            f.write("Movie")
        task.tag(path=movie_path, tags=[Tag.output(), Tag.movie()])

        yield task
        task.scratch.purge()
        task.constants._purge()


def test_transfer_l1_data(transfer_l1_data_task, mocker):
    """
    Given: A task with frames and movies tagged as output
    When: Transfering the L1 data
    Then: The task completes without errors
    """
    # Yeah, we mock a whole bunch of stuff here, but this test at least confirms that the setup to these calls is correct
    mocker.patch(
        "dkist_processing_common.tasks.mixin.metadata_store.GraphQLClient", new=FakeGQLClient
    )
    mocker.patch(
        "dkist_processing_common.tasks.mixin.globus.GlobusMixin.globus_transfer_scratch_to_object_store"
    )
    mocker.patch("dkist_processing_common.tasks.mixin.object_store.ObjectClerk.upload_object")
    mocker.patch(
        "dkist_processing_common.tasks.mixin.object_store.ObjectClerk.list_objects",
        fake_list_objects,
    )
    transfer_l1_data_task()
    assert True



"""Tag cloud manager."""
from pathlib import Path
from typing import Iterable

from redis import Redis

from dkist_processing_common._util.config import get_mesh_config


class TagDB:
    """
    Class for managing the tags database used to process and manage input data.

    Initialize a connection to a tag database whose instance is uniquely identified by the recipe_run_id and whose client connection is identified with the task_name.

    Parameters
    ----------
    recipe_run_id
        The recipe_run_id
    task_name
        The task_name
    namespace
        The namespace within the recipe_run_id
    """

    def __init__(self, recipe_run_id: int, task_name: str, namespace: str = "path"):
        self.recipe_run_id = recipe_run_id
        self.task_name = task_name
        host, port = self._db_host_port()
        self._db = self.recipe_run_id % 15
        self.namespace = f"{self.recipe_run_id}:{namespace}:"
        connection_name = f"{self.namespace}{self.task_name}"
        self.db = Redis(db=self._db, host=host, port=port, client_name=connection_name)

    @staticmethod
    def _db_host_port() -> tuple[str, int]:
        mesh_default = '{"automated-processing-scratch-inventory": {"mesh_address": "localhost", "mesh_port": 6379}}'
        mesh_config = get_mesh_config(default=mesh_default)
        return (
            mesh_config["automated-processing-scratch-inventory"]["mesh_address"],
            mesh_config["automated-processing-scratch-inventory"]["mesh_port"],
        )

    @staticmethod
    def _format_query_result(result: list[bytes]) -> set[str]:
        return {r.decode("utf8") for r in result}

    def _add_name_space(self, tags: Iterable[str] | str) -> Iterable[str] | str:
        if isinstance(tags, str):
            return self.namespace + tags
        return [f"{self.namespace}{t}" for t in tags]

    def _remove_name_space(self, raw_tags: Iterable[bytes] | bytes) -> list[str]:
        prefix_length = len(self.namespace)
        if isinstance(raw_tags, bytes):
            raw_tags = [raw_tags]
        return [str(k[prefix_length:], "UTF-8") for k in raw_tags]

    def add(self, tag: str, value: str):
        """
        Add values to a tag.

        Parameters
        ----------
        tag
            The tag
        value
            The value to be associated with the tag

        Returns
        -------
        None
        """
        tag = self._add_name_space(tag)
        self.db.sadd(tag, value)

    def clear_tag(self, tag: str) -> None:
        """
        Remove a tag from the database.

        Parameters
        ----------
        tag
            The tag to be removed

        Returns
        -------
        None
        """
        tag = self._add_name_space(tag)
        self.db.delete(tag)

    def clear_value(self, value: Path | str) -> None:
        """
        Remove a value from all tags in the DB.

        Parameters
        ----------
        value
            The value to be removed

        Returns
        -------
        None
        """
        for tag in self._namespace_keys:
            self.remove(tag, value)

    def remove(self, tag: str | bytes, value: Path | str) -> None:
        """
        Remove a value from a specific tag.

        Parameters
        ----------
        tag
            The tag
        value
            The value to be removed

        Returns
        -------
        None
        """
        if isinstance(value, Path):
            value = str(value)
        self.db.srem(tag, value)

    def any(self, tags: Iterable[str] | str) -> set[str]:
        """
        Return a set of values that match any of the tags.

        Parameters
        ----------
        tags
            The input tags

        Returns
        -------
        A set of values matching any of the tags
        """
        tags = self._add_name_space(tags)
        r = self.db.sunion(tags)
        return self._format_query_result(r)

    def all(self, tags: Iterable[str] | str) -> set[str]:
        """
        Return a set of values that match all of the tags.

        Parameters
        ----------
        tags
            The tags to be matched
        Returns
        -------
        A set of values matching all of the tags
        """
        tags = self._add_name_space(tags)
        r = self.db.sinter(tags)
        return self._format_query_result(r)

    def tags_for_value(self, value: str) -> Iterable[str]:
        """
        Return list of tags assigned to a value.

        Parameters
        ----------
        value
            The value to be matched

        Returns
        -------
        A list of tags assigned to the value
        """
        value_tags = [t for t in self._namespace_keys if self.db.sismember(t, value)]
        return self._remove_name_space(value_tags)

    def close(self):
        """
        Close the connection to the tag db.

        For use at the end of a task
        """
        self.db.close()

    @property
    def _namespace_keys(self) -> list[bytes]:
        return self.db.keys(f"{self.namespace}*")

    @property
    def tags(self) -> list[str]:
        """Return a list of all tags in this namespace."""
        return self._remove_name_space(self._namespace_keys)

    def purge(self) -> None:
        """
        Remove the database of tags.

        For use at the end of a workflow.
        """
        if keys := self._namespace_keys:
            self.db.delete(*keys)

    def __repr__(self):
        return f"TagDB(recipe_run_id={self.recipe_run_id}, task_name={self.task_name})"

    def __str__(self):
        return f"{self!r} connected to {self.db}"



"""
Framework for grouping multiple keys and values with arbitrary logic.

Defines:
    Stem -> ABC for groupings that depend on both the key and (maybe) value. Subgroups (Petals) are implied but not enforced.

    FlowerPot -> Container for Stem children (Flowers)
"""
from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from collections.abc import Hashable
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List


class FlowerPot:
    """Base class to hold multiple sets (stems) of key, value pairs."""

    def __init__(self):
        self.flowers: List[Stem] = list()

    def __iter__(self):
        return self.flowers.__iter__()

    def __len__(self):
        return self.flowers.__len__()

    def __getitem__(self, item):
        return self.flowers.__getitem__(item)

    def add_dirt(self, key: Hashable, value: Any) -> None:
        """
        Send key and value through all Flowers.

        Parameters
        ----------
        key
            The key
        value
            The value

        Returns
        -------
        None
        """
        if not isinstance(key, Hashable):
            raise TypeError(f"Type of key ({type(key)}) is not hashable")

        for flower in self.flowers:
            flower.update(key, value)


class SpilledDirt:
    """
    A custom class for when a Flower wants the FlowerPot to skip that particular key/value.

    Exists because None, False, [], (), etc. etc. are all valid Flower return values
    """


class Thorn:
    """
    Custom class to indicate that a Bud's value should not be used.

    I.e., don't pick this Bud because it's thorny. This exists for "Picky" Buds that merely perform a check and Error
    on failure. If the check passes we don't need to do anything with the Bud's value and that value should be a Thorn.
    """


class Petal:
    """
    Base class to hold a single key, value pair.

    Parameters
    ----------
    item
        The key, value pair to be added
    """

    def __init__(self, item: tuple):
        self.value = item[0]
        self.keys = item[1]

    def __repr__(self):
        return f"Petal: {{{self.value}: {self.keys}}}"


class Stem(ABC):
    """
    Base group for grouping keys via arbitrary logic on the total collection of keys and values.

    Parameters
    ----------
    stem_name
        The name to be associated with the stem
    """

    def __init__(self, stem_name: Any):
        self.stem_name = stem_name
        self.key_to_petal_dict: Dict[Hashable, Hashable] = dict()

    def update(self, key: Hashable, value: Any) -> None:
        """
        Ingest a single key/value pair.

        Parameters
        ----------
        key
            The key
        value
            The value

        Returns
        -------
        None
        """
        result = self.setter(value)
        if result is not SpilledDirt:
            self.key_to_petal_dict[key] = result

    @property
    def petals(self) -> Iterator[Petal]:
        """Return subgroups and associated keys."""
        petal_to_key_dict = defaultdict(list)
        for key, petal in self.key_to_petal_dict.items():
            petal = self.getter(key)
            petal_to_key_dict[petal].append(key)

        return (Petal(item) for item in petal_to_key_dict.items())

    @property
    def bud(self) -> Petal:
        """Just the first petal."""
        return next(self.petals)

    @abstractmethod
    def setter(self, value: Any) -> Any:
        """
        Logic to apply to a single key/value pair on ingest.

        Implemented in derived class.

        Parameters
        ----------
        value
            The value to be added

        Returns
        -------
        Any
        """
        pass

    @abstractmethod
    def getter(self, key: Hashable) -> Hashable:
        """
        Logic to apply to all ingested values when picking the Flower.

        Implemented in derived class.

        Parameters
        ----------
        key
            The key to return the value of
        Returns
        -------
        The value
        """
        pass



import pytest

from dkist_processing_common.tasks import WorkflowTaskBase


class Task(WorkflowTaskBase):
    def run(self) -> None:
        pass


@pytest.fixture
def base_task(tmp_path, recipe_run_id):
    with Task(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        yield task
    task.scratch.purge()
    task.constants._purge()


def test_apm_spans(base_task):
    """
    Given: A WorkflowTaskBase task
    When: Calling the task-specific apm_steps with weird inputs
    Then: Errors happen when they're supposed to and not when they're not supposed to
    """
    with pytest.raises(RuntimeError):
        with base_task.apm_processing_step("foo", span_type="bar"):
            pass

    with base_task.apm_task_step("foo", labels={"foo": "bar"}):
        pass



"""By-frame 214 L0 header keywords that are not instrument specific."""
from astropy.io import fits

from dkist_processing_common.parsers.l1_fits_access import L1FitsAccess


class L0FitsAccess(L1FitsAccess):
    """
    Class defining a fits access object for L0 input data.

    Parameters
    ----------
    hdu
        The input fits hdu
    name
        An optional name to be associated with the hdu
    auto_squeeze
        A boolean indicating whether to 'squeeze' out dimensions of size 1
    """

    def __init__(
        self,
        hdu: fits.ImageHDU | fits.PrimaryHDU | fits.CompImageHDU,
        name: str | None = None,
        auto_squeeze: bool = True,
    ):
        super().__init__(hdu=hdu, name=name, auto_squeeze=auto_squeeze)
        self.ip_task_type: str = self.header["IPTASK"]
        self.ip_start_time: str = self.header["DKIST011"]
        self.ip_end_time: str = self.header["DKIST012"]



"""Fault tolerant retrieval of the dkist earth location."""
from astropy import units as u
from astropy.coordinates import EarthLocation


def location_of_dkist() -> EarthLocation:
    """Return hard-coded EarthLocation of the DKIST.

    Cartesian geocentric coordinates of DKIST on Earth as retrieved from
    https://github.com/astropy/astropy-data/blob/gh-pages/coordinates/sites.json#L755
    """
    _dkist_site_info = {
        "aliases": ["DKIST", "ATST"],
        "name": "Daniel K. Inouye Solar Telescope",
        "elevation": 3067,
        "elevation_unit": "meter",
        "latitude": 20.7067,
        "latitude_unit": "degree",
        "longitude": 203.7436,
        "longitude_unit": "degree",
        "timezone": "US/Aleutian",
        "source": "DKIST website: https://www.nso.edu/telescopes/dki-solar-telescope/",
    }
    location_of_dkist = EarthLocation.from_geodetic(
        _dkist_site_info["longitude"] * u.Unit(_dkist_site_info["longitude_unit"]),
        _dkist_site_info["latitude"] * u.Unit(_dkist_site_info["latitude_unit"]),
        _dkist_site_info["elevation"] * u.Unit(_dkist_site_info["elevation_unit"]),
    )

    return location_of_dkist



"""Task(s) for the clean up tasks at the conclusion of a processing pipeline."""
import logging

from dkist_processing_common.tasks.base import WorkflowTaskBase


__all__ = ["Teardown"]


logger = logging.getLogger(__name__)


class Teardown(WorkflowTaskBase):
    """
    Changes the status of the recipe run to "COMPLETEDSUCCESSFULLY".

    Deletes the scratch directory containing all data from this pipeline run
    """

    @property
    def teardown_enabled(self) -> bool:
        """Recipe run configuration indicating if data should be removed at the end of a run."""
        return self.metadata_store_recipe_run_configuration().get("teardown_enabled", True)

    def run(self) -> None:
        """Run method for Teardown class."""
        with self.apm_task_step("Change Recipe Run to Complete Successfully"):
            self.metadata_store_change_recipe_run_to_completed_successfully()

        if not self.teardown_enabled:
            with self.apm_task_step(f"Skip Teardown"):
                logger.info(f"Data and tags for recipe run {self.recipe_run_id} NOT removed")
                return

        logger.info(f"Removing data and tags for recipe run {self.recipe_run_id}")

        with self.apm_task_step("Remove Data and Tags"):
            self.scratch.purge()

        with self.apm_task_step("Remove Constants"):
            self.constants._purge()



import json
from typing import Type

import pytest

from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.graphql import InputDatasetResponse
from dkist_processing_common.models.graphql import RecipeInstanceResponse
from dkist_processing_common.models.graphql import RecipeRunResponse
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks.teardown import Teardown


class TeardownTest(Teardown):
    def metadata_store_change_recipe_run_to_completed_successfully(self):
        pass


@pytest.fixture()
def make_mock_GQL_with_configuration():
    def class_generator(configuration: dict):
        class FakeGQLClient:
            def __init__(self, *args, **kwargs):
                pass

            @staticmethod
            def execute_gql_query(**kwargs):
                query_base = kwargs["query_base"]

                if query_base == "recipeRuns":
                    return [
                        RecipeRunResponse(
                            recipeInstanceId=1,
                            recipeInstance=RecipeInstanceResponse(
                                recipeId=1,
                                inputDataset=InputDatasetResponse(
                                    inputDatasetId=1,
                                    isActive=True,
                                    inputDatasetInputDatasetParts=[],
                                ),
                            ),
                            configuration=json.dumps(configuration),
                        ),
                    ]

            @staticmethod
            def execute_gql_mutation(**kwargs):
                ...

        return FakeGQLClient

    return class_generator


@pytest.fixture(scope="session")
def config_with_teardown_enabled() -> dict:
    return {"teardown_enabled": True}


@pytest.fixture(scope="session")
def config_with_teardown_disabled() -> dict:
    return {"teardown_enabled": False}


@pytest.fixture(scope="session")
def config_with_no_teardown() -> dict:
    return dict()


@pytest.fixture(scope="function")
def teardown_task_factory(tmp_path, recipe_run_id):
    def factory(teardown_task_cls: Type[Teardown]):
        number_of_files = 10
        tag_object = Tag.output()
        filenames = [f"file_{filenum}.ext" for filenum in range(number_of_files)]
        with teardown_task_cls(
            recipe_run_id=recipe_run_id,
            workflow_name="workflow_name",
            workflow_version="workflow_version",
        ) as task:
            task.scratch = WorkflowFileSystem(
                recipe_run_id=recipe_run_id,
                scratch_base_path=tmp_path,
            )
            task.scratch.workflow_base_path = tmp_path / str(recipe_run_id)
            for filename in filenames:
                filepath = task.scratch.workflow_base_path / filename
                filepath.touch()
                task.tag(filepath, tag_object)

            return task, filenames, tag_object

    yield factory


def test_purge_data(
    teardown_task_factory, make_mock_GQL_with_configuration, config_with_teardown_enabled, mocker
):
    """
    :Given: A Teardown task with files and tags linked to it and teardown enabled
    :When: Running the task
    :Then: All the files are deleted and the tags are removed
    """
    FakeGQLClass = make_mock_GQL_with_configuration(config_with_teardown_enabled)
    mocker.patch(
        "dkist_processing_common.tasks.mixin.metadata_store.GraphQLClient", new=FakeGQLClass
    )
    task, filenames, tag_object = teardown_task_factory(TeardownTest)
    tagged_data = list(task.read(tags=tag_object))
    for filepath in tagged_data:
        assert filepath.exists()
    task()
    for filepath in tagged_data:
        assert not filepath.exists()
    post_purge_tagged_data = list(task.read(tags=tag_object))
    assert len(post_purge_tagged_data) == 0


def test_purge_data_disabled(
    teardown_task_factory, make_mock_GQL_with_configuration, config_with_teardown_disabled, mocker
):
    """
    :Given: A Teardown task with files and tags linked to it and teardown disabled
    :When: Running the task
    :Then: All the files are not deleted and the tags remain
    """
    FakeGQLClass = make_mock_GQL_with_configuration(config_with_teardown_disabled)
    mocker.patch(
        "dkist_processing_common.tasks.mixin.metadata_store.GraphQLClient", new=FakeGQLClass
    )
    task, filenames, tag_object = teardown_task_factory(TeardownTest)
    tagged_data = list(task.read(tags=tag_object))
    for filepath in tagged_data:
        assert filepath.exists()
    task()
    for filepath in tagged_data:
        assert filepath.exists()  # still exists
    post_purge_tagged_data = list(task.read(tags=tag_object))
    assert len(post_purge_tagged_data) == len(tagged_data)


def test_purge_data_no_config(
    teardown_task_factory, make_mock_GQL_with_configuration, config_with_no_teardown, mocker
):
    """
    :Given: A Teardown task with files and tags linked and teardown not specified in the configuration
    :When: Running the task
    :Then: All the files are deleted and the tags are removed
    """
    FakeGQLClass = make_mock_GQL_with_configuration(config_with_no_teardown)
    mocker.patch(
        "dkist_processing_common.tasks.mixin.metadata_store.GraphQLClient", new=FakeGQLClass
    )
    task, filenames, tag_object = teardown_task_factory(TeardownTest)
    tagged_data = list(task.read(tags=tag_object))
    for filepath in tagged_data:
        assert filepath.exists()
    task()
    for filepath in tagged_data:
        assert not filepath.exists()
    post_purge_tagged_data = list(task.read(tags=tag_object))
    assert len(post_purge_tagged_data) == 0



"""
Global test fixtures
"""
import json
from collections import defaultdict
from datetime import datetime
from datetime import timedelta
from pathlib import Path
from random import randint
from typing import Any
from uuid import uuid4

import numpy as np
import pytest
from astropy import coordinates
from astropy.io import fits
from astropy.time import Time
from dkist_data_simulator.dataset import key_function
from dkist_data_simulator.spec122 import Spec122Dataset
from dkist_header_validator import spec122_validator
from dkist_header_validator.translator import sanitize_to_spec214_level1
from dkist_processing_pac.fitter.fitter_parameters import PolcalDresserParameters
from dkist_processing_pac.fitter.polcal_fitter import PolcalFitter
from dkist_processing_pac.input_data.drawer import Drawer
from dkist_processing_pac.input_data.dresser import Dresser
from dkist_processing_pac.optics.calibration_unit import CalibrationUnit
from dkist_processing_pac.optics.telescope import Telescope

from dkist_processing_common._util.constants import ConstantsDb
from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common._util.tags import TagDB
from dkist_processing_common.models.graphql import InputDatasetInputDatasetPartResponse
from dkist_processing_common.models.graphql import InputDatasetPartResponse
from dkist_processing_common.models.graphql import InputDatasetPartTypeResponse
from dkist_processing_common.models.graphql import InputDatasetResponse
from dkist_processing_common.models.graphql import RecipeInstanceResponse
from dkist_processing_common.models.graphql import RecipeRunResponse
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess
from dkist_processing_common.tasks import WorkflowTaskBase
from dkist_processing_common.tasks.mixin.input_dataset import InputDatasetMixin

TILE_SIZE = 64


@pytest.fixture()
def recipe_run_id():
    return randint(0, 99999)


@pytest.fixture()
def tag_db(recipe_run_id) -> TagDB:
    t = TagDB(recipe_run_id=recipe_run_id, task_name="test_tags")
    yield t
    t.purge()
    t.close()


@pytest.fixture()
def tag_db2(recipe_run_id) -> TagDB:
    """
    Another instance of a tag db in the same redis db
    """
    recipe_run_id = recipe_run_id + 15  # same db number but different namespace
    t = TagDB(recipe_run_id=recipe_run_id, task_name="test_tags2")
    yield t
    t.purge()
    t.close()


@pytest.fixture(params=[None, "use_tmp_path"])
def workflow_file_system(request, recipe_run_id, tmp_path) -> tuple[WorkflowFileSystem, int, Path]:
    if request.param == "use_tmp_path":
        path = tmp_path
    else:
        path = request.param
    wkflow_fs = WorkflowFileSystem(
        recipe_run_id=recipe_run_id,
        task_name="wkflow_fs_test",
        scratch_base_path=path,
    )
    yield wkflow_fs, recipe_run_id, tmp_path
    wkflow_fs.purge(ignore_errors=True)
    tmp_path.rmdir()
    wkflow_fs.close()


@pytest.fixture()
def constants_db(recipe_run_id) -> ConstantsDb:
    constants = ConstantsDb(recipe_run_id=recipe_run_id, task_name="test_constants")
    yield constants
    constants.purge()
    constants.close()


class CommonDataset(Spec122Dataset):
    def __init__(self):
        super().__init__(
            array_shape=(1, 10, 10),
            time_delta=1,
            dataset_shape=(2, 10, 10),
            instrument="visp",
            start_time=datetime(2020, 1, 1, 0, 0, 0, 0),
        )

        self.add_constant_key("TELEVATN", 6.28)
        self.add_constant_key("TAZIMUTH", 3.14)
        self.add_constant_key("TTBLANGL", 1.23)
        self.add_constant_key("INST_FOO", "bar")
        self.add_constant_key("DKIST004", "observe")
        self.add_constant_key("ID___005", "ip id")
        self.add_constant_key("PAC__004", "Sapphire Polarizer")
        self.add_constant_key("PAC__005", "31.2")
        self.add_constant_key("PAC__006", "clear")
        self.add_constant_key("PAC__007", "6.66")
        self.add_constant_key("PAC__008", "DarkShutter")
        self.add_constant_key("INSTRUME", "VISP")
        self.add_constant_key("WAVELNTH", 1080.0)
        self.add_constant_key("DATE-OBS", "2020-01-02T00:00:00.000000")
        self.add_constant_key("DATE-END", "2020-01-03T00:00:00.000000")
        self.add_constant_key("ID___013", "PROPOSAL_ID1")
        self.add_constant_key("PAC__002", "clear")
        self.add_constant_key("PAC__003", "on")
        self.add_constant_key("TELSCAN", "Raster")
        self.add_constant_key("DKIST008", 1)
        self.add_constant_key("DKIST009", 1)
        self.add_constant_key("BZERO", 0)
        self.add_constant_key("BSCALE", 1)


@pytest.fixture()
def complete_common_header():
    """
    A header with some common by-frame keywords
    """
    ds = CommonDataset()
    header_list = [
        spec122_validator.validate_and_translate_to_214_l0(d.header(), return_type=fits.HDUList)[
            0
        ].header
        for d in ds
    ]

    return header_list[0]


@pytest.fixture()
def complete_l1_only_header(complete_common_header):
    """
    A header with only 214 L1 keywords
    """
    complete_common_header["DAAXES"] = 1
    complete_common_header["DEAXES"] = 1
    complete_common_header["DNAXIS"] = 2
    l1_header = sanitize_to_spec214_level1(complete_common_header)

    return l1_header


class CalibrationSequenceDataset(Spec122Dataset):
    def __init__(
        self,
        array_shape: tuple[int, ...],
        time_delta: float,
        instrument="visp",
    ):
        self.num_frames_per_CS_step = 5

        # Make up a Calibration sequence. Mostly random except for two clears and two darks at start and end, which
        # we want to test
        self.pol_status = [
            "clear",
            "clear",
            "Sapphire Polarizer",
            "Fused Silica Polarizer",
            "Sapphire Polarizer",
            "clear",
            "clear",
        ]
        self.pol_theta = ["0.0", "0.0", "60.0", "60.0", "120.0", "0.0", "0.0"]
        self.ret_status = ["clear", "clear", "clear", "SiO2 SAR", "clear", "clear", "clear"]
        self.ret_theta = ["0.0", "0.0", "0.0", "45.0", "0.0", "0.0", "0.0"]
        self.dark_status = [
            "DarkShutter",
            "FieldStop (2.8arcmin)",
            "FieldStop (2.8arcmin)",
            "FieldStop (2.8arcmin)",
            "FieldStop (2.8arcmin)",
            "FieldStop (2.8arcmin)",
            "DarkShutter",
        ]

        self.num_steps = len(self.pol_theta)
        dataset_shape = (self.num_steps * self.num_frames_per_CS_step,) + array_shape[1:]
        super().__init__(
            dataset_shape,
            array_shape,
            time_delta,
            instrument=instrument,
            start_time=datetime(2020, 1, 1, 0, 0, 0),
        )
        self.add_constant_key("DKIST004", "polcal")
        self.add_constant_key("TELEVATN", 6.28)
        self.add_constant_key("ID___013", "PROPOSAL_ID1")
        self.add_constant_key("PAC__002", "clear")
        self.add_constant_key("PAC__003", "on")

    @property
    def cs_step(self) -> int:
        return self.index // self.num_frames_per_CS_step

    @key_function("PAC__004")
    def polarizer_status(self, key: str) -> str:
        return self.pol_status[self.cs_step]

    @key_function("PAC__005")
    def polarizer_angle(self, key: str) -> str:
        return str(self.pol_theta[self.cs_step])

    @key_function("PAC__006")
    def retarder_status(self, key: str) -> str:
        return self.ret_status[self.cs_step]

    @key_function("PAC__007")
    def retarder_angle(self, key: str) -> str:
        return str(self.ret_theta[self.cs_step])

    @key_function("PAC__008")
    def gos_level3_status(self, key: str) -> str:
        return self.dark_status[self.cs_step]


class NonPolCalDataset(Spec122Dataset):
    def __init__(self):
        super().__init__(
            dataset_shape=(4, 2, 2),
            array_shape=(1, 2, 2),
            time_delta=1,
            instrument="visp",
            start_time=datetime(2020, 1, 1, 0, 0, 0),
        )  # Instrument doesn't matter
        self.add_constant_key("DKIST004", "dark")  # Anything that's not polcal
        self.add_constant_key("ID___013", "PROPOSAL_ID1")
        self.add_constant_key("TELEVATN", 6.28)
        self.add_constant_key("PAC__002", "clear")
        self.add_constant_key("PAC__003", "on")
        self.add_constant_key("PAC__004", "clear")
        self.add_constant_key("PAC__005", "0.0")
        self.add_constant_key("PAC__006", "clear")
        self.add_constant_key("PAC__007", "0.0")
        self.add_constant_key("PAC__008", "DarkShutter")


@pytest.fixture(scope="session")
def grouped_cal_sequence_headers() -> dict[int, list[L0FitsAccess]]:
    ds = CalibrationSequenceDataset(array_shape=(1, 2, 2), time_delta=2.0)
    header_list = [
        spec122_validator.validate_and_translate_to_214_l0(d.header(), return_type=fits.HDUList)[
            0
        ].header
        for d in ds
    ]
    expected_cs_dict = defaultdict(list)
    for i in range(ds.num_steps):
        for j in range(ds.num_frames_per_CS_step):
            expected_cs_dict[i].append(L0FitsAccess.from_header(header_list.pop(0)))

    return expected_cs_dict


@pytest.fixture(scope="session")
def non_polcal_headers() -> list[L0FitsAccess]:
    ds = NonPolCalDataset()
    header_list = [
        spec122_validator.validate_and_translate_to_214_l0(d.header(), return_type=fits.HDUList)[
            0
        ].header
        for d in ds
    ]
    obj_list = [L0FitsAccess.from_header(h) for h in header_list]
    return obj_list


@pytest.fixture(scope="session")
def max_cs_step_time_sec() -> float:
    """Max CS step time in seconds"""
    return 20.0


class FakeGQLClient:
    def __init__(self, *args, **kwargs):
        pass

    def execute_gql_query(self, **kwargs):
        query_base = kwargs["query_base"]
        if query_base == "recipeRuns":
            return [
                RecipeRunResponse(
                    recipeInstanceId=1,
                    recipeInstance=RecipeInstanceResponse(
                        recipeId=1,
                        inputDataset=InputDatasetResponse(
                            inputDatasetId=1,
                            isActive=True,
                            inputDatasetInputDatasetParts=[
                                InputDatasetInputDatasetPartResponse(
                                    inputDatasetPart=InputDatasetPartResponse(
                                        inputDatasetPartId=1,
                                        inputDatasetPartDocument='[{"parameterName": "", "parameterValues": [{"parameterValueId": 1, "parameterValue": "[[1,2,3],[4,5,6],[7,8,9]]", "parameterValueStartDate": "1/1/2000"}]}]',
                                        inputDatasetPartType=InputDatasetPartTypeResponse(
                                            inputDatasetPartTypeName="parameters"
                                        ),
                                    )
                                ),
                                InputDatasetInputDatasetPartResponse(
                                    inputDatasetPart=InputDatasetPartResponse(
                                        inputDatasetPartId=2,
                                        inputDatasetPartDocument="""[
                                            {
                                                "bucket": "bucket_name",
                                                "object_keys": [
                                                    "key1",
                                                    "key2"
                                                ]
                                            },
                                        ]""",
                                        inputDatasetPartType=InputDatasetPartTypeResponse(
                                            inputDatasetPartTypeName="observe_frames"
                                        ),
                                    )
                                ),
                                InputDatasetInputDatasetPartResponse(
                                    inputDatasetPart=InputDatasetPartResponse(
                                        inputDatasetPartId=3,
                                        inputDatasetPartDocument="""[
                                            {
                                                "bucket": "bucket_name",
                                                "object_keys": [
                                                    "key3",
                                                    "key4"
                                                ]
                                            },
                                        ]""",
                                        inputDatasetPartType=InputDatasetPartTypeResponse(
                                            inputDatasetPartTypeName="calibration_frames"
                                        ),
                                    )
                                ),
                            ],
                        ),
                    ),
                    configuration=f'{{"tile_size": {TILE_SIZE}}}',
                ),
            ]

    @staticmethod
    def execute_gql_mutation(**kwargs):
        ...


class FakeGQLClientNoRecipeConfiguration(FakeGQLClient):
    def execute_gql_query(self, **kwargs):
        response = super().execute_gql_query(**kwargs)
        response[0].configuration = None
        return response


# All the following stuff is copied from dkist-processing-pac
def compute_telgeom(time_hst: Time):
    dkist_lon = (156 + 15 / 60.0 + 21.7 / 3600.0) * (-1)
    dkist_lat = 20 + 42 / 60.0 + 27.0 / 3600.0
    hel = 3040.4
    hloc = coordinates.EarthLocation.from_geodetic(dkist_lon, dkist_lat, hel)
    sun_body = coordinates.get_body("sun", time_hst, hloc)  # get the solar ephemeris
    azel_frame = coordinates.AltAz(obstime=time_hst, location=hloc)  # Horizon coords
    sun_altaz = sun_body.transform_to(azel_frame)  # Sun in horizon coords
    alt = sun_altaz.alt.value  # Extract altitude
    azi = sun_altaz.az.value  # Extract azimuth

    tableang = alt - azi

    return {"TELEVATN": alt, "TAZIMUTH": azi, "TTBLANGL": tableang}


class CalibrationSequenceStepDataset(Spec122Dataset):
    def __init__(
        self,
        array_shape: tuple[int, ...],
        time_delta: float,
        pol_status: str,
        pol_theta: float,
        ret_status: str,
        ret_theta: float,
        dark_status: str,
        instrument: str = "visp",
        num_mod: int = 3,
        start_time: str | datetime | None = None,
    ):
        self.num_mod = num_mod

        # Make up a Calibration sequence. Mostly random except for two clears and two darks at start and end, which
        # we want to test
        self.pol_status = pol_status
        self.pol_theta = pol_theta
        self.ret_status = ret_status
        self.ret_theta = ret_theta
        self.dark_status = dark_status
        dataset_shape = (self.num_mod,) + array_shape[1:]
        super().__init__(
            dataset_shape, array_shape, time_delta, instrument=instrument, start_time=start_time
        )
        self.add_constant_key("DKIST004", "polcal")
        self.add_constant_key("WAVELNTH", 666.0)

    @key_function("VISP_011")
    def modstate(self, key: str) -> int:
        return (self.index % self.num_mod) + 1

    @key_function("VISP_010")
    def nummod(self, key: str) -> int:
        return self.num_mod

    @key_function("PAC__004")
    def polarizer_status(self, key: str) -> str:
        return self.pol_status

    @key_function("PAC__005")
    def polarizer_angle(self, key: str) -> str:
        return "none" if self.pol_status == "clear" else str(self.pol_theta)

    @key_function("PAC__006")
    def retarter_status(self, key: str) -> str:
        return self.ret_status

    @key_function("PAC__007")
    def retarder_angle(self, key: str) -> str:
        return "none" if self.ret_status == "clear" else str(self.ret_theta)

    @key_function("PAC__008")
    def gos_level3_status(self, key: str) -> str:
        return self.dark_status

    @key_function("TAZIMUTH", "TELEVATN", "TTBLANGL")
    def telescope_geometry(self, key: str):
        return compute_telgeom(Time(self.date_obs(key), format="fits"))[key]


class InstAccess(L0FitsAccess):
    def __init__(self, hdu: fits.ImageHDU | fits.PrimaryHDU | fits.CompImageHDU):
        super().__init__(hdu, auto_squeeze=False)
        self.modulator_state = self.header["VSPSTNUM"]
        self.number_of_modulator_states = self.header["VSPNUMST"]


@pytest.fixture(scope="session")
def cs_with_correct_geometry():
    dark_status = [
        "DarkShutter",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "FieldStop (5arcmin)",
        "DarkShutter",
    ]
    ret_theta = [0, 0, 0, 0, 0, 0, 60, 120, 30, 90, 150, 0, 0, 0]
    ret_status = [
        "clear",
        "clear",
        "clear",
        "clear",
        "clear",
        "SiO2 SAR",
        "SiO2 SAR",
        "SiO2 SAR",
        "SiO2 SAR",
        "SiO2 SAR",
        "SiO2 SAR",
        "SiO2 SAR",
        "clear",
        "clear",
    ]
    pol_theta = [0, 0, 0, 60, 120, 0, 0, 0, 45, 45, 45, 45, 0, 0]
    pol_status = [
        "clear",
        "clear",
        "Sapphire Polarizer",
        "Sapphire Polarizer",
        "Sapphire Polarizer",
        "Sapphire Polarizer",
        "Sapphire Polarizer",
        "Sapphire Polarizer",
        "Sapphire Polarizer",
        "Sapphire Polarizer",
        "Sapphire Polarizer",
        "Sapphire Polarizer",
        "clear",
        "clear",
    ]
    data_shape = (1, 1, 1)
    num_steps = len(pol_theta)
    out_dict = dict()
    start_time = datetime.fromisoformat("2022-05-25T12:00:00")
    for n in range(num_steps):
        ds = CalibrationSequenceStepDataset(
            array_shape=(1, 2, 2),
            time_delta=2.0,
            pol_status=pol_status[n],
            pol_theta=pol_theta[n],
            ret_status=ret_status[n],
            ret_theta=ret_theta[n],
            dark_status=dark_status[n],
            start_time=start_time,
            num_mod=10,
        )
        header_list = [
            spec122_validator.validate_and_translate_to_214_l0(
                d.header(), return_type=fits.HDUList
            )[0].header
            for d in ds
        ]
        hdu_list = []
        for m in range(ds.num_mod):
            hdu_list.append(
                fits.PrimaryHDU(
                    data=np.ones(data_shape) * 1e3, header=fits.Header(header_list.pop(0))
                )
            )

        out_dict[n] = [InstAccess(h) for h in hdu_list]
        start_time = ds.start_time + timedelta(seconds=60)

    return out_dict, data_shape


@pytest.fixture(scope="session")
def pac_fit_mode() -> str:
    return "use_M12"


@pytest.fixture(scope="session")
def pac_init_set() -> str:
    return "OCCal_VIS"


@pytest.fixture(scope="session")
def visp_modulation_matrix() -> np.ndarray:
    # Modulation matrix for AdW's synthetic ViSP data from mod_matrix_630.out
    return np.array(
        [
            [1.0, 0.19155013, 0.80446989, -0.47479524],
            [1.0, -0.65839661, 0.68433984, 0.00466389],
            [1.0, -0.80679413, -0.16112977, 0.48234158],
            [1.0, -0.04856211, -0.56352868, 0.77578117],
            [1.0, 0.56844858, 0.03324473, 0.77289873],
            [1.0, 0.19155013, 0.80446989, 0.47479524],
            [1.0, -0.65839661, 0.68433984, -0.00466389],
            [1.0, -0.80679413, -0.16112977, -0.48234158],
            [1.0, -0.04856211, -0.56352868, -0.77578117],
            [1.0, 0.56844858, 0.03324473, -0.77289873],
        ],
        dtype=np.float64,
    )


@pytest.fixture(scope="session")
def fully_realistic_cs(
    cs_with_correct_geometry, visp_modulation_matrix, pac_fit_mode, pac_init_set
):

    cs_dict, data_shape = cs_with_correct_geometry
    dresser = Dresser()
    dresser.add_drawer(Drawer(cs_dict, skip_darks=False))
    CM = CalibrationUnit(dresser)
    TM = Telescope(dresser)
    full_params = PolcalDresserParameters(dresser, pac_fit_mode, pac_init_set)

    global_params = full_params.init_params._all_parameters[0]
    pardict = global_params.valuesdict()
    CM.load_pars_from_dict(pardict)
    TM.load_pars_from_dict(pardict)

    CM.I_sys[0] = 1e4

    # Has shape (4, N)
    S = np.sum((TM.TM @ CM.CM @ TM.M12) * CM.S_in[:, None, :], axis=2).T

    # Has shape (M, N)
    observed = visp_modulation_matrix @ S

    # Now set the "observed" value for each of the input objects
    for m in range(dresser.nummod):
        for n in range(dresser.numsteps):
            cs_dict[n][m].data *= observed[m, n] / np.mean(cs_dict[n][m].data)

    return cs_dict


@pytest.fixture(scope="session")
def fully_realistic_dresser(fully_realistic_cs):
    dresser = Dresser()
    dresser.add_drawer(Drawer(fully_realistic_cs))
    return dresser


@pytest.fixture(scope="session")
def initialized_polcal_fitter(fully_realistic_dresser, pac_init_set, pac_fit_mode) -> PolcalFitter:
    return PolcalFitter(
        local_dresser=fully_realistic_dresser,
        global_dresser=fully_realistic_dresser,
        fit_mode=pac_fit_mode,
        init_set=pac_init_set,
        fit_TM=False,
    )


class InputDatasetTask(WorkflowTaskBase, InputDatasetMixin):
    def run(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.scratch.purge()
        self.constants._purge()
        super().__exit__(exc_type, exc_val, exc_tb)


@pytest.fixture
def task_with_input_dataset(
    tmp_path, recipe_run_id, input_dataset_parts: tuple[Any, str] | list[tuple[Any, str]]
):
    if not isinstance(input_dataset_parts, list):
        input_dataset_parts = [input_dataset_parts]
    with InputDatasetTask(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        task.scratch = WorkflowFileSystem(
            recipe_run_id=recipe_run_id,
            scratch_base_path=tmp_path,
        )
        task.scratch.workflow_base_path = tmp_path / str(recipe_run_id)
        for part, tag in input_dataset_parts:
            file_path = task.scratch.workflow_base_path / Path(f"{uuid4().hex[:6]}.ext")
            file_path.write_text(data=json.dumps(part))
            task.tag(path=file_path, tags=tag)
        yield task



"""Helper functions for retrieving configurations."""
import json
from collections import defaultdict
from os import environ


def get_config(
    key: str, default: None | bool | str | dict | list | int | float = None, is_json: bool = False
) -> None | bool | str | dict | list | int | float:
    """
    Get configuration information from the specified environment variable.

    Parameters
    ----------
    key
        The environment variable from which to read the configuration.
    default
        The default value to be used if the environment variable is not defined.
    is_json
        Boolean to determine if the value should be returned as a json object or not

    Returns
    -------
    The configuration specified by the given environment variable.
    """
    value = environ.get(key, default)
    if is_json:
        return json.loads(value)
    return value


def get_mesh_config(default: str = "null") -> dict:
    """
    Get the MESH_CONFIG environment variable.

    MESH_CONFIG is an environment variable indicating how to connect to dependencies on the service mesh

    Parameters
    ----------
    default
        The default value to be used if MESH_CONFIG is not defined.

    Returns
    -------
    The value of the MECH_CONFIG environment variable
    """
    MESH_CONFIG = get_config("MESH_CONFIG", default=default, is_json=True)
    host_port_default = {"mesh_address": "localhost", "mesh_port": 5672}
    mesh_default = defaultdict(lambda: host_port_default, {})
    MESH_CONFIG = MESH_CONFIG or mesh_default
    return MESH_CONFIG



"""Spectral Line Data Structure."""


class SpectralLine:
    """
    Spectral Line Data Structure.

    Parameters
    ----------
    name
        The name of the spectral line
    wavelength
        The center wavelength of the spectral line
    wavemin
        The minimum wavelength of the spectral line
    wavemax
        The maximum wavelength of the spectral line
    """

    def __init__(self, name: str, wavelength: float, wavemin: float, wavemax: float):
        self.name = name
        self.wavelength = wavelength
        self.wavemin = wavemin
        self.wavemax = wavemax
        if not (wavemin <= wavelength <= wavemax):
            raise ValueError(
                "SpectralLine.wavemin must be smaller than SpectralLine.wavelength, which must be smaller than SpectralLine.wavemax"
            )

    @property
    def linewidth(self) -> float:
        """Return the linewidth associated with this spectral line."""
        return self.wavemax - self.wavemin

    def __repr__(self):
        return f"SpectralLine: name={self.name}, wavelength={self.wavelength}, wavemin={self.wavemin}, wavemax={self.wavemax}"


def find_associated_line(wavelength: float, lines: [SpectralLine]) -> str:
    """
    Given a list of SpectralLine objects, find the one that contains the wavelength of the frame within its range, and return its name.

    Parameters
    ----------
    wavelength
        The wavelength to match
    lines
        The list of lines

    Returns
    -------
    The name of the matching spectral line
    """
    matching_lines = []
    for line in lines:
        if line.wavemin <= wavelength <= line.wavemax:
            matching_lines.append(line)
    if len(matching_lines) == 1:
        return matching_lines[0].name
    if len(matching_lines) > 1:
        raise ValueError(
            f"{len(matching_lines)} valid spectral lines were found for this frame: {matching_lines=}, {wavelength=}"
        )
    raise ValueError(f"No valid line found for wavelength {wavelength}")



"""GraphQL Data models for the metadata store api."""
from dataclasses import dataclass
from os import environ

AUTH_TOKEN = environ.get("GQL_AUTH_TOKEN")


@dataclass
class RecipeRunMutation:
    """Dataclass used to hold a recipe run mutation record."""

    recipeRunId: int
    recipeRunStatusId: int
    authToken: str = AUTH_TOKEN


@dataclass
class RecipeRunStatusResponse:
    """Dataclass used to hold the response to a recipe run status query."""

    recipeRunStatusId: int


@dataclass
class RecipeRunStatusQuery:
    """Dataclass used to execute a recipe run status query."""

    recipeRunStatusName: str


@dataclass
class CreateRecipeRunStatusResponse:
    """Dataclass used to hold a recipe run status query response."""

    recipeRunStatus: RecipeRunStatusResponse


@dataclass
class RecipeRunStatusMutation:
    """Dataclass to hold a recipe run status mutation table."""

    recipeRunStatusName: str
    isComplete: bool
    recipeRunStatusDescription: str
    authToken: str = AUTH_TOKEN


@dataclass
class InputDatasetPartTypeResponse:
    """GraphQL response class for the input dataset part type entity."""

    inputDatasetPartTypeName: str


@dataclass
class InputDatasetPartResponse:
    """GraphQL response class for the input dataset part entity."""

    inputDatasetPartId: int
    inputDatasetPartDocument: str
    inputDatasetPartType: InputDatasetPartTypeResponse


@dataclass
class InputDatasetInputDatasetPartResponse:
    """GraphQL response class for the many-to-many join entity between input datasets and input dataset parts."""

    inputDatasetPart: InputDatasetPartResponse


@dataclass
class InputDatasetResponse:
    """Dataclass used to hold an input dataset query response."""

    inputDatasetId: int
    isActive: bool
    inputDatasetInputDatasetParts: list[InputDatasetInputDatasetPartResponse]


@dataclass
class RecipeInstanceResponse:
    """Dataclass used to hold a recipe instance query response."""

    inputDataset: InputDatasetResponse
    recipeId: int


@dataclass
class RecipeRunResponse:
    """Dataclass used to hold a recipe instance query response."""

    recipeInstance: RecipeInstanceResponse
    recipeInstanceId: int
    configuration: str = None


@dataclass
class RecipeRunQuery:
    """Dataclass used to execute a recipe run query."""

    recipeRunId: int


@dataclass
class DatasetCatalogReceiptAccountMutation:
    """
    Dataclass used to write the dataset_catalog_receipt_account record for the run.

    It sets an expected object count for a dataset so that dataset inventory creation
    doesn't happen until all objects are transferred and inventoried.
    """

    datasetId: str
    expectedObjectCount: int
    authToken: str = AUTH_TOKEN


@dataclass
class RecipeRunProvenanceMutation:
    """Dataclass used to hold a recipe run provenance mutation record."""

    inputDatasetId: int
    isTaskManual: bool
    recipeRunId: int
    taskName: str
    libraryVersions: str
    workflowVersion: str
    codeVersion: str = None
    authToken: str = AUTH_TOKEN


@dataclass
class QualityReportMutation:
    """Dataclass used to hold a quality report mutation record."""

    datasetId: str
    qualityReport: str  # JSON
    authToken: str = AUTH_TOKEN


@dataclass
class QualityReportQuery:
    """Query parameters for the metadata store endpoint qualityReports."""

    datasetId: str


@dataclass
class QualityReportResponse:
    """Query Response for the metadata store endpoint qualityReports."""

    qualityReportId: int



import json
from datetime import datetime
from typing import Any

import pytest

from dkist_processing_common.models.parameters import ParameterBase
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks import WorkflowTaskBase
from dkist_processing_common.tasks.mixin.input_dataset import InputDatasetMixin


INPUT_DATASET = [
    {
        "parameterName": "param_name",
        "parameterValues": [
            {
                "parameterValueId": 1,
                "parameterValue": json.dumps([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
                "parameterValueStartDate": "2000-01-01",
            }
        ],
    }
]

INPUT_DATASET_PARAMETERS_ONLY_NO_DATE = [
    {
        "parameterName": "param_name",
        "parameterValues": [{"parameterValueId": 1, "parameterValue": json.dumps(4)}],
    }
]

INPUT_DATASET_PARAMETERS_ONLY_TWO_VALUES = [
    {
        "parameterName": "param_name",
        "parameterValues": [
            {
                "parameterValueId": 1,
                "parameterValue": json.dumps(4),
                "parameterValueStartDate": "2020-03-13",
            },
            {
                "parameterValueId": 2,
                "parameterValue": json.dumps(6),
                "parameterValueStartDate": "1955-01-02",
            },
            {
                "parameterValueId": 3,
                "parameterValue": json.dumps(5),
                "parameterValueStartDate": "2021-12-15",
            },
        ],
    }
]

INPUT_DATASET_PARAMETERS_ONLY_TWO_VALUES_NO_DATE = [
    {
        "parameterName": "param_name",
        "parameterValues": [
            {"parameterValueId": 1, "parameterValue": json.dumps(4)},
            {
                "parameterValueId": 2,
                "parameterValue": json.dumps(6),
                "parameterValueStartDate": "1955-01-02",
            },
        ],
    }
]


class FilledParameters(ParameterBase):
    @property
    def test_parameter(self):
        return self._find_most_recent_past_value("param_name")


class ParameterScienceTask(WorkflowTaskBase, InputDatasetMixin):
    """An example of how parameters will be used in instrument repos"""

    def __init__(self, recipe_run_id: int, workflow_name: str, workflow_version: str):
        super().__init__(recipe_run_id, workflow_name, workflow_version)
        self.parameters = FilledParameters(self.input_dataset_parameters)

    def run(self) -> None:
        pass


@pytest.fixture()
def task_with_parameters(task_with_input_dataset):
    return ParameterScienceTask(
        recipe_run_id=task_with_input_dataset.recipe_run_id,
        workflow_name=task_with_input_dataset.workflow_name,
        workflow_version=task_with_input_dataset.workflow_version,
    )


@pytest.mark.parametrize(
    "input_dataset_parts",
    [pytest.param((INPUT_DATASET, Tag.input_dataset_parameters()), id="INPUT_DATASET")],
)
def test_parameters(task_with_parameters, input_dataset_parts: tuple[Any, str]):
    """
    Given: a ParameterBase subclass with populated parameters
    When: asking for a specific parameter value
    Then: the correct value is returned
    """
    assert task_with_parameters.parameters.test_parameter == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]


@pytest.mark.parametrize(
    "input_dataset_parts",
    [pytest.param((INPUT_DATASET, Tag.input_dataset_parameters()), id="INPUT_DATASET")],
)
def test_find_most_recent_date_out_of_range(
    task_with_parameters, input_dataset_parts: tuple[Any, str]
):
    """
    Given: a ParameterBase subclass with populated parameters
    When: asking for a specific parameter value at a time that is too far in the past
    Then: an error is raised
    """
    with pytest.raises(ValueError):
        task_with_parameters.parameters._find_most_recent_past_value(
            "param_name", start_date=datetime(1776, 7, 4)
        )


@pytest.mark.parametrize(
    "input_dataset_parts",
    [
        pytest.param(
            (INPUT_DATASET_PARAMETERS_ONLY_NO_DATE, Tag.input_dataset_parameters()),
            id="INPUT_DATASET_PARAMETERS_ONLY_NO_DATE",
        )
    ],
)
def test_parameters_get_no_startdate(task_with_parameters, input_dataset_parts: tuple[Any, str]):
    """
    Given: a ParameterBase subclass initialized with a parameter with no start_date
    When: asking for that specific parameter
    Then: the correct value is returned
    """
    assert task_with_parameters.parameters.test_parameter == 4


@pytest.mark.parametrize(
    "input_dataset_parts",
    [
        pytest.param(
            (INPUT_DATASET_PARAMETERS_ONLY_TWO_VALUES, Tag.input_dataset_parameters()),
            id="INPUT_DATASET_PARAMETERS_ONLY_TWO_VALUES",
        )
    ],
)
def test_find_most_recent_multiple_dates(
    task_with_parameters, input_dataset_parts: tuple[Any, str]
):
    """
    Given: a ParameterBase subclass with a parameter with multiple values
    When: asking for that specific parameter
    Then: the correct (i.e., most recent) value is returned
    """
    assert (
        task_with_parameters.parameters._find_most_recent_past_value(
            "param_name", start_date=datetime(2021, 1, 1)
        )
        == 4
    )


@pytest.mark.parametrize(
    "input_dataset_parts",
    [
        pytest.param(
            (INPUT_DATASET_PARAMETERS_ONLY_TWO_VALUES_NO_DATE, Tag.input_dataset_parameters()),
            id="INPUT_DATASET_PARAMETERS_ONLY_TWO_VALUES_NO_DATE",
        )
    ],
)
def test_parameter_get_multiple_values_no_start_date(
    task_with_parameters, input_dataset_parts: tuple[Any, str]
):
    """
    Given: a ParameterBase subclass with a parameter with multiple values, one of which has no start date
    When: asking for that specific parameter
    Then: the value with *any* date is returned
    """
    assert task_with_parameters.parameters.test_parameter == 6


@pytest.mark.parametrize(
    "input_dataset_parts",
    [pytest.param((INPUT_DATASET, Tag.input_dataset_parameters()), id="INPUT_DATASET")],
)
def test_parameters_on_task(task_with_parameters):
    """
    Given: a Task that inits a ParameterBase subclass
    When: asking for a parameter
    Then: the correct value is returned
    """
    assert task_with_parameters.parameters.test_parameter == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]



import json
from datetime import datetime
from typing import Any
from uuid import uuid4

import pytest

from dkist_processing_common.models.tags import Tag


def input_dataset_frames_part_factory(bucket_count: int = 1) -> list[dict]:
    return [
        {"bucket": uuid4().hex[:6], "object_keys": [uuid4().hex[:6] for _ in range(3)]}
        for _ in range(bucket_count)
    ]


def flatten_frame_parts(frame_parts: list[dict]) -> list[tuple[str, str]]:
    result = []
    for frame_set in frame_parts:
        for key in frame_set["object_keys"]:
            result.append((frame_set["bucket"], key))
    return result


def input_dataset_parameters_part_factory(
    parameter_count: int = 1, has_date: bool = False
) -> list[dict]:
    result = [
        {
            "parameterName": uuid4().hex[:6],
            "parameterValues": [{"parameterValueId": i, "parameterValue": json.dumps(uuid4().hex)}],
        }
        for i in range(parameter_count)
    ]
    if has_date:
        for data in result:
            data["parameterValueStartDate"] = datetime(2022, 9, 14).isoformat()[:10]
    return result


@pytest.mark.parametrize(
    "input_dataset_parts",
    [
        pytest.param((None, Tag.input_dataset_observe_frames()), id="empty"),
        pytest.param(
            (input_dataset_frames_part_factory(), Tag.input_dataset_observe_frames()),
            id="single_bucket",
        ),
        pytest.param(
            (input_dataset_frames_part_factory(bucket_count=2), Tag.input_dataset_observe_frames()),
            id="multi_bucket",
        ),
    ],
)
def test_input_dataset_observe_frames_part_document(
    task_with_input_dataset, input_dataset_parts: tuple[Any, str]
):
    """
    Given: A task with an input dataset observe frames part document tagged as such
    When: Accessing the document via the InputDatasetMixIn
    Then: The contents of the file are returned
    """
    doc_part, _ = input_dataset_parts
    task = task_with_input_dataset
    assert task.input_dataset_observe_frames_part_document == doc_part


@pytest.mark.parametrize(
    "input_dataset_parts",
    [
        pytest.param((None, Tag.input_dataset_calibration_frames()), id="empty"),
        pytest.param(
            (input_dataset_frames_part_factory(), Tag.input_dataset_calibration_frames()),
            id="single_bucket",
        ),
        pytest.param(
            (
                input_dataset_frames_part_factory(bucket_count=2),
                Tag.input_dataset_calibration_frames(),
            ),
            id="multi_bucket",
        ),
    ],
)
def test_input_dataset_calibration_frames_part_document(
    task_with_input_dataset, input_dataset_parts: tuple[Any, str]
):
    """
    Given: A task with an input dataset calibration frames part document tagged as such
    When: Accessing the document via the InputDatasetMixIn
    Then: The contents of the file are returned
    """
    doc_part, _ = input_dataset_parts
    task = task_with_input_dataset
    assert task.input_dataset_calibration_frames_part_document == doc_part


@pytest.mark.parametrize(
    "input_dataset_parts",
    [
        pytest.param((None, Tag.input_dataset_parameters()), id="empty"),
        pytest.param(
            (input_dataset_parameters_part_factory(), Tag.input_dataset_parameters()),
            id="single_param_no_date",
        ),
        pytest.param(
            (
                input_dataset_parameters_part_factory(parameter_count=2),
                Tag.input_dataset_parameters(),
            ),
            id="multi_param_no_date",
        ),
        pytest.param(
            (input_dataset_parameters_part_factory(has_date=True), Tag.input_dataset_parameters()),
            id="single_param_with_date",
        ),
        pytest.param(
            (
                input_dataset_parameters_part_factory(parameter_count=2, has_date=True),
                Tag.input_dataset_parameters(),
            ),
            id="multi_param_with_date",
        ),
    ],
)
def test_input_dataset_parameters_part_document(
    task_with_input_dataset, input_dataset_parts: tuple[Any, str]
):
    """
    Given: A task with an input dataset parameters part document tagged as such
    When: Accessing the document via the InputDatasetMixIn
    Then: The contents of the file are returned
    """
    doc_part, _ = input_dataset_parts
    task = task_with_input_dataset
    assert task.input_dataset_parameters_part_document == doc_part


@pytest.mark.parametrize(
    "input_dataset_parts",
    [
        pytest.param(
            [
                (input_dataset_frames_part_factory(), Tag.input_dataset_observe_frames()),
                (input_dataset_frames_part_factory(), Tag.input_dataset_calibration_frames()),
            ],
            id="observe1_cal1_single_bucket",
        ),
        pytest.param(
            [
                (input_dataset_frames_part_factory(), Tag.input_dataset_observe_frames()),
                (None, Tag.input_dataset_calibration_frames()),
            ],
            id="observe1_cal0_single_bucket",
        ),
        pytest.param(
            [
                (None, Tag.input_dataset_observe_frames()),
                (input_dataset_frames_part_factory(), Tag.input_dataset_calibration_frames()),
            ],
            id="observe0_cal1_single_bucket",
        ),
        pytest.param(
            [
                (None, Tag.input_dataset_observe_frames()),
                (None, Tag.input_dataset_calibration_frames()),
            ],
            id="observe0_cal0_single_bucket",
        ),
        pytest.param(
            [
                (
                    input_dataset_frames_part_factory(bucket_count=2),
                    Tag.input_dataset_observe_frames(),
                ),
                (
                    input_dataset_frames_part_factory(bucket_count=2),
                    Tag.input_dataset_calibration_frames(),
                ),
            ],
            id="observe1_cal1_multi_bucket",
        ),
        pytest.param(
            [
                (
                    input_dataset_frames_part_factory(bucket_count=2),
                    Tag.input_dataset_observe_frames(),
                ),
                (None, Tag.input_dataset_calibration_frames()),
            ],
            id="observe1_cal0_multi_bucket",
        ),
        pytest.param(
            [
                (None, Tag.input_dataset_observe_frames()),
                (
                    input_dataset_frames_part_factory(bucket_count=2),
                    Tag.input_dataset_calibration_frames(),
                ),
            ],
            id="observe0_cal1_multi_bucket",
        ),
        pytest.param(
            [
                (None, Tag.input_dataset_observe_frames()),
                (None, Tag.input_dataset_calibration_frames()),
            ],
            id="observe0_cal0_multi_bucket",
        ),
    ],
)
def test_input_dataset_frames(task_with_input_dataset, input_dataset_parts: list[tuple[Any, str]]):
    """
    Given: a task with the InputDatasetMixin
    When: getting the frames in the input dataset
    Then: it matches the frames used to create the input dataset
    """
    doc_parts = [part for part, _ in input_dataset_parts]
    task = task_with_input_dataset
    expected = []
    for part in doc_parts:
        if part:
            expected.extend(flatten_frame_parts(part))
    expected_set = set(expected)
    actual = [(frame.bucket, frame.object_key) for frame in task.input_dataset_frames]
    actual_set = set(actual)
    assert len(actual) == len(actual_set)
    assert actual_set.difference(expected_set) == set()


@pytest.mark.parametrize(
    "input_dataset_parts",
    [
        pytest.param((None, Tag.input_dataset_parameters()), id="empty"),
        pytest.param(
            (input_dataset_parameters_part_factory(), Tag.input_dataset_parameters()),
            id="single_param_no_date",
        ),
        pytest.param(
            (
                input_dataset_parameters_part_factory(parameter_count=2),
                Tag.input_dataset_parameters(),
            ),
            id="multi_param_no_date",
        ),
        pytest.param(
            (input_dataset_parameters_part_factory(has_date=True), Tag.input_dataset_parameters()),
            id="single_param_with_date",
        ),
        pytest.param(
            (
                input_dataset_parameters_part_factory(parameter_count=2, has_date=True),
                Tag.input_dataset_parameters(),
            ),
            id="multi_param_with_date",
        ),
    ],
)
def test_input_dataset_parameters(
    task_with_input_dataset, input_dataset_parts: list[tuple[Any, str]]
):
    """
    Given: a task with the InputDatasetMixin
    When: getting the parameters in the input dataset
    Then: the names of the parameters match the keys in the returned dictionary
    """
    task = task_with_input_dataset
    doc_part, _ = input_dataset_parts
    doc_part = doc_part or []  # None case parsing of expected values
    expected_parameters = {p["parameterName"]: p["parameterValues"] for p in doc_part}
    for key, values in task.input_dataset_parameters.items():
        assert key in expected_parameters
        for value in values:
            expected_values = expected_parameters[key]
            expected_value = [
                v for v in expected_values if value.parameter_value_id == v["parameterValueId"]
            ]
            assert len(expected_value) == 1
            expected_value = expected_value[0]
            assert value.parameter_value == json.loads(expected_value["parameterValue"])
            expected_date = expected_value.get("parameterValueStartDate", datetime(1, 1, 1))
            assert value.parameter_value_start_date == expected_date


@pytest.mark.parametrize(
    "input_dataset_parts",
    [
        pytest.param(
            [
                (input_dataset_frames_part_factory(), Tag.input_dataset_observe_frames()),
                (input_dataset_frames_part_factory(), Tag.input_dataset_observe_frames()),
            ],
            id="observe",
        ),
        pytest.param(
            [
                (input_dataset_frames_part_factory(), Tag.input_dataset_calibration_frames()),
                (input_dataset_frames_part_factory(), Tag.input_dataset_calibration_frames()),
            ],
            id="calibration",
        ),
        pytest.param(
            [
                (input_dataset_frames_part_factory(), Tag.input_dataset_parameters()),
                (input_dataset_frames_part_factory(), Tag.input_dataset_parameters()),
            ],
            id="params",
        ),
    ],
)
def test_multiple_input_dataset_parts(
    task_with_input_dataset, input_dataset_parts: list[tuple[Any, str]]
):
    """
    Given: a task with the InputDatasetMixin and multiple tagged input datasets
    When: reading the input dataset document
    Then: an error is raised
    """
    task = task_with_input_dataset
    with pytest.raises(ValueError):
        task.input_dataset_parameters_part_document
        task.input_dataset_observe_frames_part_document
        task.input_dataset_calibration_frames_part_document



"""Mixin for a WorkflowDataTaskBase subclass which implements interservice bus access functionality."""
import json
from os import environ
from typing import Any

from talus import DurableBlockingProducerWrapper

from dkist_processing_common._util.config import get_mesh_config


class InterserviceBusMixin:
    """Mixin for a WorkflowDataTaskBase subclass which implements interservice bus access functionality."""

    @property
    def _interservice_bus_client_retry_config(self) -> dict:
        retry_configuration = environ.get(
            "RETRY_CONFIG",
            '{"retry_delay":1,"retry_backoff":2,"retry_jitter":[1,10],"retry_max_delay":300}',
        )
        retry_configuration = json.loads(retry_configuration)
        retry_configuration["retry_jitter"] = tuple(retry_configuration["retry_jitter"])
        return retry_configuration

    @property
    def _interservice_bus_client_config(self) -> dict:
        mesh_config = get_mesh_config()
        return {
            "rabbitmq_host": mesh_config["interservice-bus"]["mesh_address"],
            "rabbitmq_port": mesh_config["interservice-bus"]["mesh_port"],
            "rabbitmq_user": environ.get("ISB_USERNAME", "guest"),
            "rabbitmq_pass": environ.get("ISB_PASSWORD", "guest"),
        }

    @staticmethod
    def _interservice_bus_parse_bindings(messages: list) -> list[dict[str, str]]:
        # @message_class decorated messages only
        # transform in to a single list
        all_bindings = []
        for m in messages:
            all_bindings += m.binding()
        # dedup
        result = []
        for binding in all_bindings:
            if binding not in result:
                result.append(binding)
        return result

    def interservice_bus_publish(self, messages: list[Any] | object):
        """Publish messages on the interservice bus."""
        if not isinstance(messages, list):
            messages = [messages]
        # @message_class decorated messages only
        bindings = self._interservice_bus_parse_bindings(messages=messages)
        with DurableBlockingProducerWrapper(
            producer_queue_bindings=bindings,
            publish_exchange="master.direct.x",
            **self._interservice_bus_client_config,
            **self._interservice_bus_client_retry_config,
        ) as producer:
            for message in messages:
                producer.publish_message(message)



"""Pre-made flower that produces tag based on a single header key."""
from dkist_processing_common.models.flower_pot import Stem
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess


class SingleValueSingleKeyFlower(Stem):
    """
    Flower that just passes through a single header value.

    Parameters
    ----------
    tag_stem_name
        The tag stem name
    metadata_key
        The metadata key
    """

    def __init__(self, tag_stem_name: str, metadata_key: str):
        super().__init__(stem_name=tag_stem_name)
        self.metadata_key = metadata_key

    def setter(self, fits_obj: L0FitsAccess):
        """
        Set the value.

        Parameters
        ----------
        fits_obj
            The input fits object

        Returns
        -------
        The value associated with the metadata key for this object
        """
        return getattr(fits_obj, self.metadata_key)

    def getter(self, key):
        """
        Get the value.

        Parameters
        ----------
        key
            The input metadata key

        Returns
        -------
        The value associated with the metadata key for this object
        """
        return self.key_to_petal_dict[key]



"""Mixin class supporting the generation of the quality reports.

To improve readability the top-level mixin, `QualityMixin`, contains only base functionality and the separate
metrics are grouped into sub-mixins. To protect a user, this mixin-on-mixin stack is hidden in protected modules
and only the top-level mixin (`QualityMixin`) is exposed.
"""
from ._base import QualityMixin
